/*
 * transaction.hpp
 *
 * Copyright (C) 2009 Tim Marston <edam@waxworlds.org>
 *
 * This file is part of sqlitepp (hereafter referred to as "this program").
 * See http://www.waxworlds.org/edam/software/sqlitepp for more information.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef TRANSACTION_HPP_
#define TRANSACTION_HPP_


#include <boost/utility.hpp>


namespace sqlite
{


class database;


class transaction
	:
	private boost::noncopyable
{
//______________________________________________________________________________
//                                                                 instantiation
public:

	/**
	 * Constructor that provides a database upon which to act
	 * @param database a database
	 */
	transaction(
		database &database );

	virtual ~transaction() throw( );

protected:

	/**
	 * Constructor that provides a way for deriving classes to override the SQL
	 * executed in beginning and rolling-back a transaction during construction
	 * and destruction.
	 * @param database a database
	 * @param begin_sql the SQL statement used to begin the transaction
	 * @param rollback_sql the SQL statement used to rollback the transaction,
	 * 		or an empty string if the default is to be used.
	 */
	transaction(
		database &database,
		const std::string &begin_sql,
		const std::string &rollback_sql = "" );

//______________________________________________________________________________
//                                                              public interface

	/**
	 * Commit the transaction
	 */
	virtual void commit();

	/**
	 * Rollback the transaction
	 */
	void rollback();

//______________________________________________________________________________
//                                                                implementation
protected:

	/** the database */
	database &_database;

	/** the SQL used to rollback the transaction, or empty to use default */
	std::string _rollback_sql;

};


////////////////////////////////////////////////////////////////////////////////


class exclusive_transaction
	:
	public transaction
{
//______________________________________________________________________________
//                                                                 instantiation
public:

	/**
	 * Constructor that provides a database upon which to act
	 * @param database a database
	 */
	exclusive_transaction(
		database &database );

};


}


#endif /* TRANSACTION_HPP_ */
