/*
 * statement.cpp
 *
 * Copyright (C) 2009 Tim Marston <edam@waxworlds.org>
 *
 * This file is part of sqlitepp (hereafter referred to as "this program").
 * See http://www.waxworlds.org/edam/software/sqlitepp for more information.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sqlitepp/statement.hpp>
#include <sqlitepp/exception.hpp>
#include <sqlitepp/database.hpp>
#include <string.h>
#include <iomanip>


const sqlite::_null_t sqlite::null = { };

const sqlite::_exec_t sqlite::exec = { };


sqlite::_set_index_t sqlite::set_index(
	unsigned int index )
{
	_set_index_t t = { index };
	return t;
}


sqlite::statement::statement(
	database &database,
	const std::string &sql )
	:
	_database( database ),
	_handle( NULL ),
	_bind_index( 1 )
{
	int error_code = prepare( sql );
	if( error_code != SQLITE_OK ) throw sqlite_error( error_code );
}


sqlite::statement::~statement() throw( )
{
	finalize();
}


int sqlite::statement::prepare(
	const std::string &sql )
{
	finalize();
	return sqlite3_prepare_v2( _database._handle, sql.c_str(),
		sql.length() + 1, &_handle, NULL );
}


int sqlite::statement::step()
{
	return sqlite3_step( _handle );
}


int sqlite::statement::reset()
{
	return sqlite3_reset( _handle );
}


int sqlite::statement::clear_bindings()
{
	_bind_index = 1;
	return sqlite3_clear_bindings( _handle );
}


int sqlite::statement::bind_static(
	unsigned int index,
	const char *value,
	unsigned int value_length )
{
	return sqlite3_bind_text( _handle, index, value, value_length,
		SQLITE_STATIC );
}


int sqlite::statement::bind_static(
	unsigned int index,
	const char *value )
{
	return bind_static( index, value, strlen( value ) );
}


int sqlite::statement::bind_static(
	unsigned int index,
	const std::string &value )
{
	return bind_static( index, value.c_str(), value.length() );
}


int sqlite::statement::bind_null(
	unsigned int index )
{
	return sqlite3_bind_null( _handle, index );
}


int sqlite::statement::bind_static(
	const std::string &name,
	const char *value,
	unsigned int value_length )
{
	unsigned int index = sqlite3_bind_parameter_index( _handle, name.c_str() );
	return bind_static( index, value, value_length );
}


int sqlite::statement::bind_static(
	const std::string &name,
	const char *value )
{
	return bind_static( name, value, ::strlen( value ) );
}


int sqlite::statement::bind_static(
	const std::string &name,
	const std::string &value )
{
	return bind_static( name, value.c_str(), value.length() );
}


int sqlite::statement::bind_null(
	const std::string &name )
{
	unsigned int index = sqlite3_bind_parameter_index( _handle, name.c_str() );
	return bind_null( index );
}


int sqlite::statement::finalize()
{
	int error_code = SQLITE_OK;

	if( _handle ) {
		error_code = sqlite3_finalize( _handle );
		_handle = NULL;
	}

	return error_code;
}


template< >
sqlite::statement& sqlite::statement::operator << < sqlite::_null_t >(
	sqlite::_null_t )
{
	int error_code = bind_null( _bind_index );
	if( error_code != SQLITE_OK ) throw sqlite_error( error_code );
	_bind_index++;
	return *this;
}

template< >
sqlite::statement& sqlite::statement::operator << < sqlite::_exec_t >(
	sqlite::_exec_t )
{
	int error_code = step();
	if( error_code != SQLITE_DONE )
		if( error_code == SQLITE_ROW )
			throw sqlite_error( "statement returned results" );
		else
			throw sqlite_error( error_code );
	return *this;
}

template< >
sqlite::statement& sqlite::statement::operator << < sqlite::_set_index_t >(
	sqlite::_set_index_t t )
{
	_bind_index = t._index;
	return *this;
}
